@@ -1,6 +1,6 @@
 /*
  *   stunnel       TLS offloading and load-balancing proxy
- *   Copyright (C) 1998-2016 Michal Trojnara <Michal.Trojnara@stunnel.org>
+ *   Copyright (C) 1998-2017 Michal Trojnara <Michal.Trojnara@stunnel.org>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -295,6 +295,8 @@ char *log_id(CLI *c) {
         if(!tid) /* currently USE_FORK */
             tid=stunnel_process_id();
         return str_printf("%lu", tid);
+    case LOG_ID_PROCESS:
+        return str_printf("%lu", stunnel_process_id());
     }
     return str_dup("error");
 }
